"""Onyx Client URL helper."""
from typing import Optional, Any

import aiohttp

from onyx_client.configuration.configuration import Configuration
from onyx_client.utils.const import API_URL, API_HEADERS, API_VERSION
from onyx_client.utils.response import check


class UrlHelper:
    """URL helper for performing requests against the HELLA.ONYX API."""

    def __init__(self, config: Configuration, client_session: aiohttp.ClientSession):
        """Initialize the helper.

        config: the ONYX configuration
        client_session: the aiohttp client session to use"""
        self.config = config
        self.client_session = client_session

    @property
    def _headers(self) -> dict:
        """Get all common headers."""
        return {"Authorization": f"Bearer {self.config.access_token}", **API_HEADERS}

    def _base_url(self, with_api: bool = True) -> str:
        """Get the API base URL for this ONYX.CENTER.

        with_api: append the API version to the URL"""
        api = f"{API_URL}/box/{self.config.fingerprint}/api"
        if with_api:
            api = f"{api}/{API_VERSION}"
        return api

    def _url(self, path: str = "", with_api: bool = True) -> str:
        """Get the request URL.

        path: the URL path
        with_api: append the API version to the URL"""
        return f"{self._base_url(with_api=with_api)}{path}"

    async def perform_get_request(
        self, path: str, with_api: bool = True
    ) -> Optional[Any]:
        """Perform a GET request.

        path: the URL path
        with_api: append the API version to the URL"""
        async with self.client_session.get(
            self._url(path, with_api=with_api), headers=self._headers
        ) as response:
            if not check(response):
                return None
            return await response.json()

    async def perform_delete_request(self, path: str) -> Optional[Any]:
        """Perform a DELETE request.

        path: the URL path"""
        async with self.client_session.delete(
            self._url(path), headers=self._headers
        ) as response:
            if not check(response):
                return None
            return await response.json()

    async def perform_post_request(self, path: str, data: dict) -> Optional[Any]:
        """Perform a POST request.

        path: the URL path
        data: the data to POST to the API"""
        async with self.client_session.post(
            self._url(path), json=data, headers=self._headers
        ) as response:
            if not check(response):
                return None
            return await response.json()

    async def start_stream(self, path: str):
        """Start a stream and returns the value if it's not empty.

        path: the URL path"""
        async with self.client_session.get(
            self._url(path),
            headers=self._headers,
            timeout=aiohttp.ClientTimeout(
                total=0, connect=0, sock_connect=0, sock_read=0
            ),
        ) as response:
            if not check(response):
                yield None
                return
            async for message in response.content:
                cleaned_message = str(message.strip(), "UTF-8").strip()
                if len(cleaned_message) > 0:
                    yield cleaned_message



"""Weather class."""
from onyx_client.data.device_mode import DeviceMode
from onyx_client.data.numeric_value import NumericValue
from onyx_client.device.device import Device
from onyx_client.enum.device_type import DeviceType


class Weather(Device):
    """A ONYX controlled weather station."""

    def __init__(
        self,
        identifier: str,
        name: str,
        device_type: DeviceType,
        device_mode: DeviceMode,
        actions: list,
        wind_peak: NumericValue = None,
        sun_brightness_peak: NumericValue = None,
        sun_brightness_sink: NumericValue = None,
        air_pressure: NumericValue = None,
        humidity: NumericValue = None,
        temperature: NumericValue = None,
    ):
        """Initialize the weather device.

        identifier: the device identifier
        name: the device name
        device_type: the device type
        actions: a list of actions the device supports
        wind_peak: the maximum wind speed in the last 15 minutes
        sun_brightness_peak: the maximum brightness in the last 15 minutes
        sun_brightness_sink: the minimum brightness in the last 15 minutes
        air_pressure: the absolute air pressure
        humidity: the relative humidity
        temperature: the temperature of the station"""
        super(Weather, self).__init__(
            identifier, name, device_type, device_mode, actions
        )
        self.wind_peak = wind_peak
        self.sun_brightness_peak = sun_brightness_peak
        self.sun_brightness_sink = sun_brightness_sink
        self.air_pressure = air_pressure
        self.humidity = humidity
        self.temperature = temperature

    def update_with(self, update):
        """Update the device with an update patch.

        update: the update patch"""
        super().update_with(update)

        self.wind_peak = (
            self.wind_peak if update.wind_peak is None else update.wind_peak
        )
        self.sun_brightness_peak = (
            self.sun_brightness_peak
            if update.sun_brightness_peak is None
            else update.sun_brightness_peak
        )
        self.sun_brightness_sink = (
            self.sun_brightness_sink
            if update.sun_brightness_sink is None
            else update.sun_brightness_sink
        )
        self.air_pressure = (
            self.air_pressure if update.air_pressure is None else update.air_pressure
        )
        self.humidity = self.humidity if update.humidity is None else update.humidity
        self.temperature = (
            self.temperature if update.temperature is None else update.temperature
        )

    @staticmethod
    def keys() -> list:
        """Get the list of keys specific to the device type."""
        return [
            "wind_peak",
            "sun_brightness_peak",
            "sun_brightness_sink",
            "air_pressure",
            "humidity",
            "temperature",
        ]



"""Animation Keyframe of Onyx devices."""


class AnimationKeyframe:
    """The representation of an animation keyframe."""

    def __init__(self, interpolation: str, delay: int, duration: float, value: int):
        """Initialize the animation keyframe.

        interpolation: the interpolation algorithm
        delay: the delay of the animation
        duration: the duration of the animation
        value: the value to interpolate to"""
        self.interpolation = interpolation
        self.delay = delay
        self.duration = duration
        self.value = value

    @staticmethod
    def create(properties: dict):
        """Create an animation keyframe from properties.

        properties: the properties of the device"""
        if properties is None:
            return None

        return AnimationKeyframe(
            properties.get("interpolation", None),
            properties.get("delay", None),
            properties.get("duration", None),
            properties.get("value", None),
        )

    def __str__(self) -> str:
        return f"AnimationKeyframe(interpolation={self.interpolation}, delay={self.delay}, duration={self.duration}, value={self.value})"

    def __eq__(self, other):
        if isinstance(self, other.__class__):
            return (
                self.interpolation == other.interpolation
                and self.delay == other.delay
                and self.duration == other.duration
                and self.value == other.value
            )
        return False



"""Shutter class."""
from onyx_client.data.device_mode import DeviceMode
from onyx_client.data.numeric_value import NumericValue
from onyx_client.device.device import Device
from onyx_client.enum.device_type import DeviceType


# not mapped properties:
# - auto_calibration
# - heart_beat_enabled
# - sun_guard_lower_action
# - sun_guard_lower_angle
# - sun_guard_lower_position
# - sun_guard_lower_threshold
# - sun_guard_upper_action
# - sun_guard_upper_angle
# - sun_guard_upper_position
# - sun_guard_upper_threshold
# - system_state
# - wind_guard_enabled
# - wind_peak_threshold
class Shutter(Device):
    """A ONYX controlled shutter device."""

    def __init__(
        self,
        identifier: str,
        name: str,
        device_type: DeviceType,
        device_mode: DeviceMode,
        actions: list,
        target_position: NumericValue = None,
        target_angle: NumericValue = None,
        actual_angle: NumericValue = None,
        actual_position: NumericValue = None,
    ):
        """Initialize the shutter device.

        identifier: the device identifier
        name: the device name
        device_type: the device type
        device_mode: the mode the device can operate as
        actions: a list of actions the device supports
        target_position: the target position of the shutter
        target_angle: the target angle of the shutter
        actual_position: the actual position of the shutter
        actual_angle: the actual angle of the shutter"""
        super(Shutter, self).__init__(
            identifier, name, device_type, device_mode, actions
        )
        self.target_position = target_position
        self.target_angle = target_angle
        self.actual_angle = actual_angle
        self.actual_position = actual_position

    def update_with(self, update):
        """Update the device with an update patch.

        update: the update patch"""
        super().update_with(update)

        self.target_position = (
            self.target_position
            if update.target_position is None
            else update.target_position
        )
        self.target_angle = (
            self.target_angle if update.target_angle is None else update.target_angle
        )
        self.actual_angle.update_with(update.actual_angle)
        self.actual_position.update_with(update.actual_position)

    @staticmethod
    def keys() -> list:
        """Get the list of keys specific to the device type."""
        return [
            "target_position",
            "target_angle",
            "actual_angle",
            "actual_position",
        ]



"""Onyx Client API configuration."""


class Configuration:
    """The API connection configuration."""

    def __init__(self, fingerprint: str, access_token: str):
        """Initialize the configuration.

        fingerprint: the ONYX.CENTER device fingerprint
        access_token: the access token to use"""
        self.fingerprint = fingerprint
        self.access_token = access_token



"""Switch class."""
from onyx_client.data.device_mode import DeviceMode
from onyx_client.device.device import Device
from onyx_client.enum.device_type import DeviceType


class Switch(Device):
    """A ONYX controlled switch device."""

    def __init__(self, identifier: str, name: str, device_type: DeviceType):
        """Initialize the switch device.

        identifier: the device identifier
        name: the device name
        device_type: the device type"""
        super(Switch, self).__init__(
            identifier,
            name,
            device_type,
            DeviceMode(device_type),
            list(),
        )

    def update_with(self, update):
        """Update the device with an update patch.

        update: the update patch"""
        super().update_with(update)

    @staticmethod
    def keys() -> list:
        """Get the list of keys specific to the device type."""
        return []



"""Device Command for an Onyx device."""
from onyx_client.enum.action import Action
from onyx_client.exception.invalid_command import InvalidCommandException


class DeviceCommand:
    """The representation of a device command."""

    def __init__(
        self,
        properties: dict = None,
        action: Action = None,
        valid_from: int = None,
        best_before: int = None,
    ):
        """Initialize the command.

        properties: the properties to send
        action: the action to perform
        valid_from: the start date of when the action can be triggered
        bedt_before: deadline of the action validity"""
        if (properties is None and action is None) or (
            properties is not None and action is not None
        ):
            raise InvalidCommandException("COMMAND_NO_PROPERTIES_OR_ACTION")

        self.properties = properties
        self.action = action
        self.valid_from = valid_from
        self.best_before = best_before

    def __str__(self) -> str:
        return f"DeviceCommand(data={self.data()})"

    def data(self) -> dict:
        """Get the dict representation."""
        data = {}
        if self.properties is not None:
            data = {**data, "properties": {**self.properties}}
        if self.action is not None:
            data = {**data, "action": self.action.name.lower()}
        if self.valid_from is not None:
            data = {**data, "valid_from": self.valid_from}
        if self.best_before is not None:
            data = {**data, "best_before": self.best_before}
        return data



"""Group class."""


class Group:
    """A ONYX controlled group."""

    def __init__(
        self,
        identifier: str,
        name: str,
        devices: list,
    ):
        """Initialize the group.

        identifier: the group identifier
        name: the group name
        devices: the list of devices belonging to the group"""
        self.identifier = identifier
        self.name = name
        self.devices = devices

    def __eq__(self, other):
        if isinstance(self, other.__class__):
            return self.identifier == other.identifier
        return False



"""Boolean Values of Onyx devices."""


class BooleanValue:
    """The representation of a boolean value."""

    def __init__(self, value: bool, read_only: bool):
        """Initialize the boolean value.

        value: the value
        read_only: set if the value is read only"""
        self.value = value
        self.read_only = read_only

    @staticmethod
    def create(properties: dict):
        """Create a boolean value from properties.

        properties: the properties of the device"""
        if properties is None:
            return None

        return BooleanValue(
            properties.get("value", "false") == "true",
            properties.get("read_only", "false") == "true",
        )

    def __str__(self) -> str:
        return f"BooleanValue(value={self.value})"

    def __eq__(self, other):
        if isinstance(self, other.__class__):
            return self.value == other.value and self.read_only == other.read_only
        return False



"""Date Information (date, time, timezone, ...) class."""


class DateInformation:
    """Container for all date related information of the ONYX.CENTER."""

    def __init__(self, time: float, timezone: str, timezone_offset: int):
        """Initialize the date information.

        time: the value
        timezone: the timezone of the device
        timezone_offset: the offset of the timezone"""
        self.time = time
        self.timezone = timezone
        self.timezone_offset = timezone_offset



"""Device Actions of Onyx devices."""
from enum import Enum, auto


class Action(Enum):
    """The actions supported by Onyx."""

    CLOSE = auto()
    OPEN = auto()
    STOP = auto()
    TILT_DOWN = auto()
    TILT_UP = auto()
    WINK = auto()
    LIGHT_ON = auto()
    LIGHT_OFF = auto()

    def string(self) -> str:
        """Get the string representation for the Onyx API."""
        return self.name.lower()

    @staticmethod
    def convert(lower: str):
        """Get the action from the Onyx API's type information.

        lower: the lower case action identifier"""
        try:
            return Action[lower.upper()]
        except KeyError:
            return None



"""Device class."""
from onyx_client.data.device_mode import DeviceMode
from onyx_client.enum.device_type import DeviceType
from onyx_client.exception.update_exception import UpdateException


class Device:
    """A ONYX controlled device."""

    def __init__(
        self,
        identifier: str,
        name: str,
        device_type: DeviceType,
        device_mode: DeviceMode,
        actions: list,
    ):
        """Initialize the device.

        identifier: the device identifier
        name: the device name
        device_type: the device type
        actions: a list of actions the device supports"""
        self.identifier = identifier
        self.name = name
        self.device_type = device_type
        self.device_mode = device_mode
        self.actions = actions

    def __str__(self):
        return (
            f"Device(id={self.identifier}, name={self.name}, type={self.device_type})"
        )

    def __eq__(self, other):
        if isinstance(self, other.__class__):
            return self.identifier == other.identifier
        return False

    def update_with(self, update):
        """Update the device with an update patch.

        update: the update patch"""
        if not self == update:
            raise UpdateException("ID_NOT_EQUAL")
        self.name = self.name if update.name is None else update.name
        self.device_type = (
            self.device_type if update.device_type is None else update.device_type
        )
        self.device_mode = (
            self.device_mode if update.device_mode is None else update.device_mode
        )
        self.actions = (
            self.actions
            if update.actions is None or len(update.actions) == 0
            else update.actions
        )



"""Device Mode for an Onyx device."""
from onyx_client.enum.device_type import DeviceType


class DeviceMode:
    """The device mode for the device."""

    def __init__(self, mode: DeviceType, values: list = None):
        """Initialize the device mode.

        mode: the device type
        values: a list of valid values"""
        self.mode = mode
        self.values = values
